<?php
/**
 * Copyright © 2015 FiloBlu. All rights reserved.
 */

namespace FiloBlu\Flow\Model;

use \FiloBlu\Flow\Model\Channel\In\Images;

class Processor
{

    const MAX_IMAGE_PROCESS_RETRY_COUNT = 3;
    const CHANNEL_TYPE_IMAGE_CODE = 'image';

    /**
     * @var Monolog\Logger
     */
    protected $_logger;
    protected $_lockFile = array();
    protected $_map = array();
    protected $_connector = array();
    protected $_parser = array();
    protected $_config = array();
    protected $_helper_data;
    protected $_channels;
    protected $channels = array();
    /**
     * @var Inboundflow
     */
    protected $_inboundflow;
    protected $_imageflow;
    protected $_inbound_items = array();
    protected static $_has_exception_in_process = false;


    /**
     * Processor constructor.
     * @param \Magento\Framework\Model\Context $context
     * @param \Magento\Framework\Registry $registry
     * @param \FiloBlu\Flow\Helper\Data $helper_data
     * @param \FiloBlu\Flow\Helper\LoggerProvider $loggerProvider
     * @param Channel $channels
     * @param Inboundflow $inboundflow
     * @param Imagesflow $imageFlow
     * @param \Magento\Framework\App\Config\ScopeConfigInterface $scopeConfig
     * @param \Magento\Framework\Stdlib\DateTime\DateTime $date
     * @param array $data
     */
    public function __construct(
        \Magento\Framework\Model\Context $context,
        \Magento\Framework\Registry $registry,
        \FiloBlu\Flow\Helper\Data $helper_data,
        \FiloBlu\Flow\Helper\LoggerProvider $loggerProvider,
        \FiloBlu\Flow\Model\Channel $channels,
        \FiloBlu\Flow\Model\Inboundflow $inboundflow,
        \FiloBlu\Flow\Model\Imagesflow $imageFlow,
        \Magento\Framework\App\Config\ScopeConfigInterface $scopeConfig,
        \Magento\Framework\Stdlib\DateTime\DateTime $date,
        array $data = []
    )
    {
        $this->_logger = $loggerProvider->getLogger();
        $this->_channels = $channels;
        $this->_scopeConfig = $scopeConfig;
        $this->_helper_data = $helper_data;
        $this->date = $date;
        $this->_inboundflow = $inboundflow;
        $this->_imageflow = $imageFlow;

        $attributesToCopy = \Magento\Framework\App\ObjectManager::getInstance()->create('FiloBlu\Flow\Model\System\ProductAttributes')->getAttributesToCopy()->getColumnValues('attribute_code');
        $registry->register('attribute_configurable_copy', $attributesToCopy);

        $this->init();
    }

    /**
     *
     */
    private function init()
    {
        $this->initConfig();
        $this->setChannels();
        $this->setInboundflowItems();
    }

    /**
     *
     */
    public function initConfig()
    {

        $collection = $this->_channels->getCollection();

        foreach ($collection as $channel) {

            $config_data = json_decode($channel->getData('data'));

            $channel_name = strtolower($channel->getData('name'));
            $channel_model_class = $this->_helper_data->getChannelModels($config_data->config->flow);
            $channel_type = $channel->getData('type');

            $this->_config['channels']->{$channel_type}[$channel_name] = $channel_model_class;
        }
    }

    /**
     * @param string $type
     * @return $this
     */
    public function setChannels($type = "in")
    {

        $channels = $this->_config['channels'];

        // $this->_logger->info('Channels - ' . print_r($channels, true));

        foreach ((Array)$channels->$type as $channel_name => $channel_model_class) {

            $channel_class = '\FiloBlu\Flow\Model\Channel' . "\\" . ucfirst($type) . "\\" . ucfirst($channel_model_class);

            $model = false;

            try {
                $objectManager = \Magento\Framework\App\ObjectManager::getInstance();
                $model = $objectManager->create($channel_class);
                $model->load($channel_name);
            } catch (\Exception $e) {
                $this->_logger->info($e->getMessage());
            }

            if ($model) {
                $this->channels[$channel_name] = $model;
            } else {
                //throw new \Exception('Error in channel '.$type.' '.$channel_model_class.' config');
                $this->_logger->info('Error in channel ' . $type . ' ' . $channel_model_class . ' config. CHANNEL SKIPPED');
            }
        }
        return $this;
    }

    /**
     * @return array
     */
    public function setInboundflowItems()
    {

        $this->_logger->info(__METHOD__);

        if (!$this->_inbound_items) {

            $_inbound_collection = $this->_inboundflow->getCollection();

            foreach ($_inbound_collection as $inbound) {
                $this->_inbound_items[] = $inbound->getData('name');
            }
        }

        return $this->_inbound_items;
    }

    /**
     * @param int $bus
     * @throws \Exception
     */
    public function fetchChannels($bus = 0)
    {

        // $this->_logger->info(__METHOD__);

        //preleva l'elenco dei channel e li passa al connettore
        $channels = $this->getChannels();

        if (!count($channels)) {
            throw new \Exception('no channel configured');
        }

        // il connettore restituisce le risorse da interrogare
        foreach ($channels as $channel_name => $channel) {

            // controllo che $bus sia diverso dal default e sia uguale a quello passato
            if ($bus && $channel->getConfig('bus') != $bus) {
                continue;
            }

            //chiama il connector in base alla configurazione
            $list = $channel->rlist();
            if (count($list) && $list !== false) {
                $this->populateModel($channel, $channel_name, $list);
            }
        }
    }

    /**
     * @param $channel
     * @param $channel_name
     * @param $list
     */
    protected function populateModel($channel, $channel_name, $list)
    {

        switch ($channel->getConfig('type')) {

            case self::CHANNEL_TYPE_IMAGE_CODE:

                $model = $this->_inboundflow;
                $model->setName("images_" . date('Ymd_Hmi') . ".csv");
                $model->setFlow($channel->getConfig('flow'));
                $model->setType($channel->getConfig('type'));
                $model->setChannel($channel_name);
                $model->setRetry(0);
                $model->setStatus('found');
                $model->setPriority($channel->getConfig('priority'));
                try {
                    $model->save();
                    $lid = $model->getId();
                } catch (\Exception $e) {
                    $this->_logger->info($e->getMessage());
                }

                if (!$lid) {
                    //handle error here
                    return false;
                }

                $model = $this->_imageflow;

                $model = $this->_imageflow;

                foreach ($list as $item) {
                    $model->setId(null);
                    $model->setMetaFile($lid);
                    $model->setImageName($item);
                    $model->setImageStatus('found');
                    $model->setData('image_queue_time', date("Y-m-d H:i:s"));
                    try {
                        $model->save();
                    } catch (\Exception $e) {
                        $this->_logger->info($e->getMessage());
                    }
                }
                break;

            default:

                foreach ($list as $item) {

                    // se il file è già presente nella tabella inbound, non deve essere reimportato
                    if (in_array($item, $this->_inbound_items)) continue;

                    $model = $this->_inboundflow;
                    $model->setName($item);
                    $model->setFlow($channel->getConfig('flow'));
                    $model->setType($channel->getConfig('type'));
                    $model->setChannel($channel_name);
                    $model->setRetry(0);
                    $model->setStatus('found');
                    $model->setPriority($channel->getConfig('priority'));
                    try {
                        $model->save();
                    } catch (\Exception $e) {
                        $this->_logger->info($e->getMessage());
                    }
                }
        }
    }

    /**
     *
     */
    public function receiveData()
    {

        // $this->_logger->info(__METHOD__);
        $channel = '';

        // preleva i file da ricevere

        $files = $this->_inboundflow->getCollection();

        $files->addFieldToFilter('status', 'found');

        foreach ($files as $file) {
            try {

                if ($file->getStatus() == 'receiving')
                    continue;

                $file->setStatus('receiving')->save();

                $channel = $this->channels[$file->getData('channel')];

                if (!($channel instanceof \FiloBlu\Flow\Model\Channel\In\AbstractModel)) {
                    $message = 'Error in channel ' . $channel->model . ' config during receiving';
                    $file->setLog($message)->save();
                    throw new \Exception($message);
                }

                $received = $channel->receiveFile($file);

                if ($received) {
                    $file->setStatus('received')->save();
                } else {
                    $file->setStatus('error')->save();
                }

            } catch (\Exception $e) {

                $file->setLog($e->getMessage())->save();

            }
        }

    }

    /**
     * @param int|null $file_id
     */
    public function parse($file_id = null)
    {
        $files = \Magento\Framework\App\ObjectManager::getInstance()->create('FiloBlu\Flow\Model\ResourceModel\Inboundflow\Collection');
        $files->addFieldToFilter('status', 'received');

        foreach ($files as $file) {

            if ($file->getStatus() == 'parsing') {
                continue;
            }

            $file->setStatus('parsing')->save();

            $channel = $this->channels[$file->getData('channel')];

            if ($channel) {
                try {
                    $parsed = $channel->parse($file);
                    if ($parsed)
                        $file->setStatus('parsed')->save();
                    else
                        $file->setStatus('error')->save();
                } catch (\Exception $e) {
                    $file->setLog($e->getMessage())->save();
                    $file->setStatus('error')->save();
                }
            } else {
                $file->setLog('No channel found with name ' . $file->getData('channel'))->save();
            }
        }
    }

    /**
     * oggetto channel from name
     */
    public function getChannel($name)
    {
        return (isset($this->channels[$name])) ? $this->channels[$name] : null;
    }

    /**
     * all channels from config, oggetti di tipo Channel
     */
    public function getChannels()
    {
        return $this->channels;
    }

    /**
     * @param string $type
     * @param string $priority
     */
    public function processNormal($type = 'file', $priority = 'normal')
    {

        $files_model = $this->_inboundflow->getCollection()->addFieldToFilter('status', 'parsed');

        $files = $files_model->getItems();

        usort($files, function ($a, $b) {
            $_process_order = array(
                'From\Product' => 0,
                'From\Eavs' => 1,
                'From\Stock' => 2,
                'From\Multistock' => 3,
                'From\Price' => 4
            );
            $a['val'] = isset($_process_order[$a->getFlow()]) ? $_process_order[$a->getFlow()] : 5;
            $b['val'] = isset($_process_order[$b->getFlow()]) ? $_process_order[$b->getFlow()] : 5;
            return ($a['val'] < $b['val']) ? -1 : 1;
        });

        foreach ($files as $file) {
            $this->_logger->info($file->getFlow());
            $file_type = $file->getType();

            try {

                switch ($file_type) {
                    case self::CHANNEL_TYPE_IMAGE_CODE:
                        $this->processImages($file);
                        break;
                    default:
                        $this->processFile($file);
                }

            } catch (\Exception $e) {
                $file->setStatus('parsed');
                $file->setLog(__METHOD__ . ': ' . $e->getMessage() . "\n\n" . $e->getTraceAsString())->save();
            }
        }

    }

    /**
     * @param $file
     * @return mixed
     */
    public function processImages($file)
    {
        if($file->getStatus() === Images::STATUS_ERROR) {
            return;
        }

        $file->setStatus('processing')->save();

        /** @var \FiloBlu\Flow\Model\Channel\In\Images $model */
        $model = \Magento\Framework\App\ObjectManager::getInstance()->create(Images::class);

        $result = $model->insertData($file);

        if ($result) {
            $file->setStatus(Images::STATUS_PROCESSED)->save();
            return;
        }

        $total = $model->getImageToProcessCount();
        $errors = $model->getProcessedImageErrorCount();

        $retry = $file->getRetry();
        $log = $file->getLog() . "Process #{$retry}: Image to process {$total} ({$errors} fault)\n";

        $file->setRetry($retry + 1);

        if ($retry >= self::MAX_IMAGE_PROCESS_RETRY_COUNT) {
            $file->setLog("Unable to complete the flow after {$retry} tries. Look at flow images table for more details.\n\n" . $log);
            $file->setStatus(Images::STATUS_ERROR);
        } else {
            $file->setStatus(Images::STATUS_PARSED);
            $file->setLog($log);
        }

        $file->save();

        return;
    }

    /**
     * @param $file
     * @return bool
     */
    public function processFile($file)
    {
        $file->setStatus('processing')->save();
        $model = \Magento\Framework\App\ObjectManager::getInstance()->create('FiloBlu\Flow\Model\\' . $file->getFlow());

        if (!$model) {
            return false;
        }

        try {
            $model->processFileRows($file);
            $file->setComment($model->getErrorMessage());
        } catch (\Exception $e) {
            $file->setStatus('parsed');
            $this->_logger->info($e->getMessage());
            $file->setLog($e->getMessage())->save();
        }

        return $file->setStatus('processed')->save();
    }

    /**
     * @throws \Exception
     */
    public function clean()
    {
        $days = (int)$this->_scopeConfig->getValue('filoblu_flow/flow_clean/flow_clean_days', \Magento\Store\Model\ScopeInterface::SCOPE_STORE);

        if (!$days) {
            throw new \Exception("no days for clean specified");
        }

        $files_model = $this->_inboundflow->getCollection()
            //->addFieldToFilter('type', $type)
            ->addFieldToFilter('status', array('processed', 'error'))
            ->addFieldToFilter('receiving_time', array("lt" => new \Zend_Db_Expr("DATE_SUB(CURDATE(), INTERVAL $days DAY)")));

        $connection = $model->getResource()->getConnection();

        $files = $files_model->getItems();

        foreach ($files as $file) {
            // $this->_logger->info($file->getFlow());
            $file_type = $file->getType();
            try {

                switch ($file_type) {
                    case self::CHANNEL_TYPE_IMAGE_CODE:
                        //$this->processImages($file);
                        break;
                    default:

                        $sql_last = "select max(id) as id from " . $model->getResource()->getMainTable() . " where channel = " . '"' . $file->getChannel() . '"';
                        $result = $connection->fetchRow($sql_last);

                        if ($result["id"] != $file->getId()) {
                            $this->cleanFile($file);
                        }
                }

            } catch (\Exception $e) {
                $file->setLog($e->getMessage())->save();
            }
        }
    }

    /**
     * @param $file
     * @return bool
     */
    public function cleanFile($file)
    {
        $model = \Magento\Framework\App\ObjectManager::getInstance()->create('FiloBlu\Flow\Model\\' . $file->getFlow());

        if (!$model) {
            return false;
        }

        try {
            $model->clean($file);
            $file->setStatus('cleaned')->save();
        } catch (\Exception $e) {
            $this->_logger->info($e->getMessage());
            $file->setLog($e->getMessage())->save();
        }
    }

}
